/**
 * Copyright (C) 2011 Ovea <dev@ovea.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ovea.resource.web;

import com.mycila.xmltool.CallBack;
import com.mycila.xmltool.ValidationResult;
import com.mycila.xmltool.XMLDoc;
import com.mycila.xmltool.XMLTag;
import com.ovea.resource.web.util.Resource;
import com.ovea.resource.web.util.ResourceException;

import javax.servlet.ServletContext;
import java.io.*;
import java.net.JarURLConnection;
import java.net.URL;
import java.util.Arrays;
import java.util.Deque;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

final class XmlBasedWebResourceProvider {

    private final ServletContext servletContext;
    private final String descriptor;
    private volatile VirtualResources virtualResources;
    private boolean debug;

    XmlBasedWebResourceProvider(ServletContext servletContext, String descriptor) {
        this.servletContext = servletContext;
        this.descriptor = descriptor;
    }

    public void setDebug(boolean debug) {
        this.debug = debug;
    }

    public VirtualResources get() {
        if (virtualResources == null || debug) {
            VirtualResources virtualResources = new VirtualResources();
            virtualResources.setDebug(debug);
            Enumeration<URL> resources;
            try {
                resources = Thread.currentThread().getContextClassLoader().getResources(descriptor);
            } catch (IOException e) {
                throw new RuntimeException(e.getMessage(), e);
            }
            while (resources.hasMoreElements()) {
                append(virtualResources, resources.nextElement());
            }
            this.virtualResources = virtualResources;
        }
        return virtualResources;
    }

    private void append(final VirtualResources virtualResources, URL url) {
        ValidationResult result = XMLDoc.from(url, false).validate(Resource.from("classpath:com/ovea/resource/web/virtual-resources.xsd").toURL());
        if (result.hasError())
            throw new IllegalStateException("Document " + url + " is not valid: " + Arrays.deepToString(result.getErrorMessages()), result.getErrors()[0]);
        XMLTag tag = XMLDoc.from(url, true);
        tag.forEach("map", new CallBack() {
            @Override
            public void execute(XMLTag mapTag) {
                String newBasedir = mapTag.getAttribute("virtual-path");
                boolean optional = mapTag.hasAttribute("optional") ? Boolean.valueOf(mapTag.getAttribute("optional")) : false;
                Resource to = Resource.from(mapTag.getAttribute("to"));
                if (to.isFile()) {
                    Deque<File> files = new LinkedList<File>();
                    try {
                        String baseDir = to.toFile().getCanonicalFile().getAbsolutePath();
                        files.add(new File(baseDir));
                        while (!files.isEmpty()) {
                            File file = files.pop();
                            if (file.isDirectory())
                                for (File child : file.listFiles())
                                    files.addFirst(child);
                            else {
                                try {
                                    String relativePath = file.getCanonicalPath().substring(baseDir.length()).replace("\\", "/");
                                    if (!relativePath.startsWith("/"))
                                        relativePath = relativePath + "/";
                                    VirtualResource virtualResource = virtualResources.addVirtualResource(newBasedir + relativePath);
                                    WebResource wr = virtualResource.addWebResource(Resource.from(file));
                                    wr.setOptional(optional);
                                    if (mapTag.hasAttribute("caching"))
                                        virtualResource.setCaching(Long.parseLong(mapTag.getAttribute("caching")));
                                    if (mapTag.hasAttribute("gzip"))
                                        virtualResource.setGzip(Boolean.parseBoolean(mapTag.getAttribute("gzip")));
                                } catch (IOException e) {
                                    if (!optional)
                                        throw e;
                                }
                            }
                        }
                    } catch (IOException e) {
                        if (!optional)
                            throw new RuntimeException(e.getMessage(), e);
                    }
                } else {
                    try {
                        String path = to.toURL().getPath();
                        path = path.substring(path.indexOf("!") + 2);
                        JarURLConnection jarCon = (JarURLConnection) to.toURL().openConnection();
                        JarFile jarFile = jarCon.getJarFile();
                        Enumeration<JarEntry> e = jarFile.entries();
                        while (e.hasMoreElements()) {
                            JarEntry jarEntry = e.nextElement();
                            if (!jarEntry.isDirectory() && jarEntry.getName().startsWith(path)) {
                                VirtualResource virtualResource = virtualResources.addVirtualResource(newBasedir + jarEntry.getName().substring(path.length()));
                                WebResource wr = virtualResource.addWebResource(Resource.fromData(read(jarFile.getInputStream(jarEntry))));
                                wr.setOptional(optional);
                                if (mapTag.hasAttribute("caching"))
                                    virtualResource.setCaching(Long.parseLong(mapTag.getAttribute("caching")));
                                if (mapTag.hasAttribute("gzip"))
                                    virtualResource.setGzip(Boolean.parseBoolean(mapTag.getAttribute("gzip")));
                            }
                        }
                    } catch (ResourceException e1) {
                        if (!optional)
                            throw e1;
                    } catch (IOException e1) {
                        throw new RuntimeException(e1.getMessage(), e1);
                    }
                }
            }
        });
        tag.forEach("resource", new CallBack() {
            public void execute(XMLTag resTag) {
                final VirtualResource virtualResource = virtualResources.addVirtualResource(resTag.getAttribute("virtual-path"));
                if (resTag.hasAttribute("contentType"))
                    virtualResource.setContentType(resTag.getAttribute("contentType"));
                if (resTag.hasAttribute("caching"))
                    virtualResource.setCaching(Long.parseLong(resTag.getAttribute("caching")));
                if (resTag.hasAttribute("gzip"))
                    virtualResource.setGzip(Boolean.parseBoolean(resTag.getAttribute("gzip")));
                resTag.forEachChild(new CallBack() {
                    public void execute(XMLTag locTag) {
                        WebResource webResource = virtualResource.addWebResource(buidResource(locTag));
                        if (locTag.hasAttribute("minimize"))
                            webResource.setMinimize(Boolean.parseBoolean(locTag.getAttribute("minimize")));
                        if (locTag.hasAttribute("optional"))
                            webResource.setOptional(Boolean.parseBoolean(locTag.getAttribute("optional")));
                    }
                });
            }
        });
    }

    private Resource buidResource(XMLTag res) {
        switch (WebResourceLocation.from(res.getCurrentTagName())) {
            case CLASSPATH: {
                return Resource.from("classpath:" + res.getText());
            }
            case WEB: {
                String path = servletContext.getRealPath(res.getText());
                File file = path == null ? null : new File(path);
                return Resource.from(file);
            }
            case CODE: {
                return Resource.fromData(res.getCDATAorText());
            }
        }
        throw new IllegalArgumentException(res.getCurrentTagName());
    }

    private static byte[] read(InputStream in) throws IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        try {
            in = new BufferedInputStream(in);
            byte[] buffer = new byte[8096];
            int c;
            while ((c = in.read(buffer)) != -1)
                baos.write(buffer, 0, c);
            return baos.toByteArray();
        } finally {
            in.close();
        }
    }

}